import shutil
import string
import subprocess as sp
import sys
import tempfile
from datetime import datetime
from functools import wraps
from pathlib import Path
from pprint import pprint
from time import time
from typing import Dict, List, Optional

import httpx
import trio
import typer
from halo import Halo

from invisible_hand.config import app_context
from ..ensures import ensure_config_exists, ensure_gh_token
from ..utils.github_scanner import get_github_endpoint_paged_list_async
from ..utils.google_student import Gstudents
from .shared_options import opt_all_yes, opt_gh_org, opt_github_token

# We will use {source_repo}/{source_tmpl_issue} as the template to give students feedback

# The Pull request we made would fetch the first issue which has the same title
# as patch_branch to be the content


def measure_time(f):
    @wraps(f)
    def wrap(*args, **kw):
        ts = time()
        result = f(*args, **kw)
        te = time()
        return result, te - ts

    return wrap


def announce_grade(
    homework_prefix: str = typer.Argument(..., help="prefix of the target homework"),
    feedback_source_repo: Optional[str] = typer.Option(
        None, show_default=True, help="Repo contains students' feedbacks"
    ),
    only_id: Optional[str] = typer.Option(default=None, help="only id to announce"),
    token: Optional[str] = opt_github_token,
    org: str = opt_gh_org,
    dry: str = typer.Option(
        False, "--dry", help="dry run, do not publish result to the remote"
    ),
    yes: bool = opt_all_yes,
):
    """Announce student grades to each hw repo"""
    ensure_config_exists()

    def fallback(val, fallback_value):
        return val if val else fallback_value

    # Handle default value manually because we'll change our config after app starts up
    token: str = fallback(token, app_context.config.github.personal_access_token)
    org: str = fallback(org, app_context.config.github.organization)
    feedback_source_repo: str = fallback(
        feedback_source_repo, app_context.config.announce_grade.feedback_source_repo
    )

    ensure_gh_token(token)
    if not (yes or typer.confirm(f"Add annouce_grade to {org}/{homework_prefix}?")):
        raise typer.Abort()

    # TODO: use logging lib to log messages
    spinner = Halo(stream=sys.stderr)

    student_feedback_title = f"Grade for {homework_prefix}"

    gstudents = Gstudents()
    feedback_vars = gstudents.left_join(homework_prefix)

    # Clone feedback repo & set needed variables
    cur = Path(".")

    for d in cur.glob("feedback-tmp-*"):
        shutil.rmtree(d)
    spinner.info("delete dated folder")

    root_folder = Path(
        tempfile.mkdtemp(
            prefix="feedback-tmp-{}-".format(datetime.now().strftime("%b%d%H%M%S")),
            dir=".",
        )
    )
    spinner.succeed(f"Create tmp folder {root_folder}")

    feedback_repo_path = root_folder / "feedbacks"

    spinner.info(f"cloning feeback source repo : {feedback_source_repo}")
    _, t = measure_time(sp.run)(
        [
            "git",
            "clone",
            f"https://github.com/{org}/{feedback_source_repo}.git",
            feedback_repo_path.name,
        ],
        cwd=root_folder,
    )
    spinner.succeed(
        f"cloning feeback source repo : {feedback_source_repo} ... {t:4.2f} sec"
    )
    client = httpx.AsyncClient(
        headers=httpx.Headers(
            {
                "User-Agent": "GitHubClassroomUtils/1.0",
                "Authorization": "token " + token,
                # needed for the check-suites request
                "Accept": "application/vnd.github.antiope-preview+json",
            }
        )
    )

    hw_path = feedback_repo_path / homework_prefix / "reports"

    # generate feedbacks
    fbs, t = measure_time(gen_feedbacks)(homework_prefix, hw_path, feedback_vars)
    spinner.succeed(f"Generate content for feedbacks ... {t:5.3f} sec")

    # handle only_id
    if only_id:
        try:
            # detect possible buggy condition
            info = gstudents.get_student(only_id)
        except RuntimeError as e:
            print(" *=" * 30)
            print("Warning!")
            print(e)
            return
        only_repo_name = get_hw_repo_name(homework_prefix, info["github_handle"])
        fbs = list(filter(lambda fb: fb["repo_name"] == only_repo_name, fbs))

    async def push_to_remote(feedback_title, feedbacks):
        # push to remote
        async def push_feedback(fb):
            request_body = {"title": feedback_title, "body": fb["value"]}
            try:
                issue_num = await find_existing_issue(
                    client, org, fb["repo_name"], feedback_title
                )
            except BaseException:
                print(f'error on {fb["repo_name"]}')
                return
            if issue_num:
                request_body["state"] = "open"  # reopen issue
                url = f"https://api.github.com/repos/{org}/{fb['repo_name']}/issues/{issue_num}"
                await edit_issue_async(client, url, issue_num, request_body)
            else:
                url = f"https://api.github.com/repos/{org}/{fb['repo_name']}/issues"
                await create_issue_async(client, url, request_body)
            print(f'success {fb["repo_name"]}')

        async with trio.open_nursery() as nursery:
            for fb in feedbacks:
                nursery.start_soon(push_feedback, fb)

    # print out target repos
    print("repo to announce grade:")
    pprint([fb["repo_name"] for fb in fbs])

    if dry:
        spinner.succeed("DRYRUN: skip push to remote")
    else:
        if typer.confirm("Do you want to continue?", default=False):
            _, t = measure_time(trio.run)(push_to_remote, student_feedback_title, fbs)
            spinner.succeed(f"Push feedbacks to remote ... {t:5.2f} sec")
        else:
            spinner.warn("You refused to publish to remote")

    spinner.succeed("finished announce grade")
    return


def get_hw_repo_name(hw_prefix, gh_handle):
    return f"{hw_prefix}-{gh_handle}"


def gen_feedbacks(hw_prefix, fb_root_path, fb_vars) -> List[Dict]:
    # substitude with template info
    result = []
    for vars in fb_vars:
        feedback = {"repo_name": None, "value": None}
        # generate names for student h.w. repos
        student_id, handle = vars["student_id"], vars["github_handle"]
        repo_name = get_hw_repo_name(hw_prefix, handle)
        feedback["repo_name"] = repo_name
        # fetch each students feedback template file by id
        fb_tmpl_file = fb_root_path / f"{student_id}.md"
        if fb_tmpl_file.exists():
            # substitude value inside feedback template
            with open(fb_tmpl_file, "r") as tmpl_file:
                fb_tmpl = string.Template(tmpl_file.read())
                feedback["value"] = fb_tmpl.safe_substitute(vars)

        if feedback["repo_name"] and feedback["value"]:
            result.append(feedback)
    return result


async def find_existing_issue(client, org, repo_name, issue_title) -> Optional[int]:
    # TODO: remove argument: org
    # print(f'finding existing issue... :{repo_name}/{issue_title}')
    issues = await get_github_endpoint_paged_list_async(
        client, f"repos/{org}/{repo_name}/issues", state="all"
    )
    # TODO: deal with repo not exist case
    if len(issues) != 0:
        try:
            issue = next(i for i in issues if i["title"] == issue_title)
            return int(issue["number"])
        except:
            pass
    return None


async def edit_issue_async(client, url, issue_number, requst_body):
    res = await client.patch(url, json=requst_body)
    if res.status_code != 200:
        print(f"failed on editing issue: {url}")
        try:
            print(f"{res.json()['errors'][0]['message']}")
        except:
            pass


async def create_issue_async(client, url, request_body):
    res = await client.post(url, json=request_body)
    if res.status_code != 201:
        print(f"failed on {url}")
        try:
            print(f"{res.json()['errors'][0]['message']}")
        except:
            pass


if __name__ == "__main__":
    announce_grade()



from typing import Optional

import trio
import typer
from tqdm import tqdm

from invisible_hand.config import app_context
from ..ensures import ensure_config_exists, ensure_gh_token
from ..utils.github_entities import Team
from ..utils.github_scanner import query_matching_repos
from .shared_options import opt_all_yes, opt_dry, opt_gh_org, opt_github_token


def grant_read_access(
    # @TODO: add autocompletion for option
    # https://typer.tiangolo.com/tutorial/options/autocompletion/
    hw_title: str = typer.Argument(
        default=..., metavar="📝hw", help="target homework to grant access to"
    ),
    team: Optional[str] = typer.Option(
        None, "--team", help="team-slug of the Teaching teams", show_default=True,
    ),
    dry: bool = opt_dry,
    yes: bool = opt_all_yes,
    token: Optional[str] = opt_github_token,
    org: Optional[str] = opt_gh_org,
):
    """Make TAs being able to read all homework repos

    "team-slug": the field auto generated by github.

    For example: the team-slug of "2019 Teaching-team" would be "2019_teaching-team".
    """
    ensure_config_exists()
    print("start script")

    def fallback(val, fallback_value):
        return val if val else fallback_value

    # Handle default value manually because we'll change our config after app starts up
    token: str = fallback(token, app_context.config.github.personal_access_token)
    org: str = fallback(org, app_context.config.github.organization)
    team: str = fallback(team, app_context.config.grant_read_access.reader_team_slug)

    ensure_gh_token(token)
    if not (
        yes or typer.confirm(f"Grant read access of {org}/{hw_title} to {org}/{team}?")
    ):
        raise typer.Abort()

    teaching_team = Team(org, team, token, dry)

    repos = query_matching_repos(
        org, github_repo_prefix=f"{hw_title}-", github_token=token, verbose=True
    )
    repo_names = [r["name"] for r in repos]

    # show repos to operate on
    ncols = 3
    cols = [repo_names[i::ncols] for i in range(ncols)]
    print("repos to operate on:")
    for a, b, c in zip(*cols):
        print(f"  {a:<30}{b:<30}{c:<}")

    builder = pbar_builder()
    builder.set_config(total=len(repo_names))
    with builder.build(desc="fired") as fired, builder.build(desc="returned") as returned:

        async def subscribe_to_repo(team: Team, repo_name: str):
            res = await team.add_team_repository_async(repo_name, permission="pull")
            if res.status_code != 204:
                print(res)
            returned.update(1)

        async def async_github():
            async with trio.open_nursery() as nursery:
                for repo_name in repo_names:
                    if not dry:
                        nursery.start_soon(subscribe_to_repo, teaching_team, repo_name)
                    fired.update(1)
                    fired.refresh()
                return

        trio.run(async_github)


class pbar_builder:
    def __cover_dict(self, src, dst):
        for k, v in src.items():
            dst[k] = v

    def __init__(self):
        self.config = {}
        self.next_po = 0

    def set_config(self, **kwargs):
        self.__cover_dict(kwargs, self.config)

    def build(self, **kwargs):
        cfg = {}
        self.__cover_dict(self.config, cfg)
        self.__cover_dict(kwargs, cfg)
        cfg["position"] = self.next_po
        self.next_po += 1
        return tqdm(**cfg)



from typing import Optional

import typer
from pydantic.error_wrappers import ValidationError

from .config import app as config_typer
from .config import app_context
from .scripts import (
    add_students,
    announce_grade,
    crawl_classroom,
    event_times,
    grant_read_access,
    patch_project,
)

CONTEXT_SETTINGS = {"help_option_names": ["-h", "--help"]}
app = typer.Typer(
    context_settings=CONTEXT_SETTINGS,
    help="Manage your classroom under Github organization.",
)


@app.callback()
def main(
    custom_base: Optional[str] = typer.Option(
        None, "--custom-base", help="Use custom base folder for configs"
    )
):
    config_manager = app_context.config_manager

    if custom_base:
        config_manager.change_base_folder(custom_base)
        typer.echo(f"Using custom base-folder: {config_manager.get_base_path()}")

    # populate actual config
    if config_manager.config_path.exists():
        try:
            app_context.config = config_manager.read_config()
        except ValidationError:
            typer.echo("[Warning] Corrupted config file")


app.command("add-students")(add_students)
app.command("grant-read-access")(grant_read_access)
app.command("event-times")(event_times)
app.command("patch-project")(patch_project)
app.command("crawl-classroom")(crawl_classroom)
app.command("announce-grade")(announce_grade)

app.add_typer(config_typer, name="config")



import atexit
import shutil
import sys
from pathlib import Path
from typing import Optional

import requests
import typer
from bs4 import BeautifulSoup
from halo import Halo
from selenium import webdriver
from selenium.webdriver.chrome.options import Options
from selenium.webdriver.common.by import By
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.support.ui import WebDriverWait

from invisible_hand.config import app_context
from invisible_hand.ensures import ensure_config_exists
from invisible_hand.errors import ERR_CHROME_DRIVER_NOT_INSTALLED

spinner = Halo(stream=sys.stderr)

# must reside in root folder
# change the extension version for newer versions
EXTENSION_FILE_VERSION = ""
EXTENSION_FILE = "chrome-ghclassroom-waiter{EXTENSION_FILE_VERSION}.crx"


def cache_extension_path() -> Path:
    """ Download and cache chrome extension from github
    """
    extension_path: Path = app_context.config_manager.base_folder / EXTENSION_FILE
    url: str = f"https://github.com/ianre657/invisible-hand/raw/master/{EXTENSION_FILE}"
    if not extension_path.exists():
        response = requests.get(url, allow_redirects=True)
        print(response.json)
        if response.status_code == 200:
            open(extension_path, "wb").write(response.content)
        else:
            raise RuntimeError("Cannot download file from {url}")
    return extension_path


class Condition_logined:
    def __init__(self):
        pass

    def __call__(self, driver):
        try:
            # will throw an exception if target not found
            _ = driver.find_element_by_id("login_field")
            # print("I can still see login field")
            return False
        except:
            # print("no more login field found")
            return True


# TODO:
#  - error handling
#     - deadline is not due
#
#  - show progress
#     - how many repos in the page that are not loaded
#  - show progress bar in each page
#
#  - report statistics
#     - how many repos
#     - how many students have submitted their hw
#     - how many students didn't submitted their hw and their info


BaseURL = "https://classroom.github.com"


def crawl_classroom(
    hw_title=typer.Argument(..., help="homework submission to crawl"),
    output: str = typer.Argument(..., help="filename to output your result"),
    classroom_id: Optional[str] = typer.Option(None, show_default=True),
    login: Optional[str] = typer.Option(None),
    passwd: Optional[str] = typer.Option(
        default=None, help="password for your github classroom"
    ),
):
    """Get student's submissions from Github Classroom
    """
    ensure_config_exists()

    def fallback(val, fallback_value):
        return val if val else fallback_value

    # Handle default value manually because we'll change our config after app starts up
    classroom_id: str = fallback(
        classroom_id, app_context.config.crawl_classroom.classroom_id
    )
    login: str = fallback(login, app_context.config.crawl_classroom.login)

    spinner.start()
    if output is not None:
        if Path(output).exists():
            spinner.info(f"Overwrite existing file: {output}")

    spinner.text = "Initializing Chrome driver"
    # We can only use headless=False
    # Because I use my chrome extension helping my web crawling
    # but Google doesn't want to support Chrome extensions in headless mode
    # See https://bugs.chromium.org/p/chromium/issues/detail?id=706008#c5 for more information
    driver = createDriver(headless=False)
    atexit.register(lambda: driver.quit)

    # gh classroom assigment page
    spinner.text = "Opening GitHub Classroom"
    driver.get(BaseURL + "/classrooms/{}/assignments/{}".format(classroom_id, hw_title))

    # user id
    input_id = driver.find_element_by_id("login_field")
    input_id.send_keys(login)

    # user password
    input_passwd = driver.find_element_by_id("password")
    spinner.text = "Login..."

    if passwd is not None:
        input_passwd.send_keys(passwd)
        driver.find_element_by_name("commit").click()
        try:
            _ = WebDriverWait(driver, 0).until(Condition_logined())
            spinner.text = "Login to GitHub Classroom"
            spinner.succeed()
        except:
            spinner.fail("Invalid password")
            sys.exit(1)

    try:
        logined = Condition_logined()
        _ = WebDriverWait(driver, 20).until(logined)
    except:
        spinner.fail("Browser timed out")
        sys.exit(1)

    # check if error
    submitted_infos = []
    not_submitted_list = []
    page_loaded = 0
    while True:
        spinner.text = "Loading page {}".format(page_loaded + 1)
        spinner.start()
        wait_for_page_load(driver)
        page_loaded += 1
        spinner.succeed("Page {} loaded".format(page_loaded))
        soup = BeautifulSoup(driver.page_source, "lxml")
        next_btn = soup.find("a", "next")

        li = soup.find_all("div", "assignment-repo-list-item")
        for i in li:
            user_id = i.find("a", "assignment-repo-github-url").find("h3").contents[0]
            links_to_commit = i.find("a", attrs={"aria-label": "View Submission"})

            if links_to_commit is not None:
                commit_url = links_to_commit["href"]
                commit_hash = commit_url.rsplit("/", 1)[-1]
                submitted_str = "{}-{}:{}".format(hw_title, user_id, commit_hash[0:7])
                submitted_infos.append(submitted_str)
                # print(submitted_str)
            else:
                not_submitted_list.append(f"{hw_title}-{user_id}")

        if next_btn == None:
            break
        driver.get(BaseURL + next_btn["href"])
    spinner.succeed("Success")

    ostream = sys.stdout
    if output is not None:
        ostream = open(output, "w")
        atexit.register(lambda: ostream.close)

    for i in submitted_infos:
        print(i, end=" ", file=ostream)
    print("", file=ostream)

    if len(not_submitted_list) != 0:
        print("Students not submitted:", file=sys.stderr)
        print(not_submitted_list, file=sys.stderr)


def createDriver(headless=True) -> webdriver:

    options = Options()
    options.headless = headless
    options.add_argument("--window-size=1920,1200")

    options.add_extension(cache_extension_path())

    cdriver_path = shutil.which("chromedriver")
    if cdriver_path is None:
        raise ERR_CHROME_DRIVER_NOT_INSTALLED
    driver = webdriver.Chrome(options=options, executable_path=cdriver_path)
    return driver


def wait_for_page_load(driver):
    try:
        _ = WebDriverWait(driver, 30).until(
            # This id is generated by my chrome extension
            EC.presence_of_element_located(
                (By.ID, "chrome-extension-classroom-waiter-loaded")
            )
        )
    except:
        driver.close()
        raise ("Timeout: cannot load assignment page")


if __name__ == "__main__":
    crawl_classroom()



import sys
from typing import List, Optional, Tuple

import requests
import typer
from halo import Halo
from rich.columns import Columns
from rich.console import Console
from rich.panel import Panel

from invisible_hand.config import app_context
from ..ensures import ensure_config_exists, ensure_gh_token
from ..utils.github_entities import Team
from ..utils.github_scanner import github_headers
from .shared_options import opt_all_yes, opt_dry, opt_gh_org, opt_github_token


def print_table(data):
    """Prints formatted data in columns"""
    users = [Panel(user, expand=True) for user in data]
    columns = Columns(users, equal=True)
    Console().print(columns)


# Use a "team slug"
# for example: "2019 Students-hello" -> "2019_students-hello"


def check_is_github_user(github_id, github_token) -> bool:
    res = requests.get(
        f"https://api.github.com/users/{github_id}", headers=github_headers(github_token),
    )
    return res.status_code == 200


class SafetyActor:
    """ Rexport functions to respect dry run option.
    Which would not make the actual call if `dry==true`
    """

    def __init__(self, dry: bool):
        self._dry = dry

    def _safe_trigger(self, func, *args, **kwargs):
        __tracebackhide__ = True
        triggered, val = False, None
        if not self._dry:
            val = func(*args, **kwargs)
            triggered = True
        else:
            pass
            # typer.echo("safe enabled, not fire ")
        return triggered, val

    def ensure_gh_token(self, token: str):
        triggered, val = self._safe_trigger(ensure_gh_token, token=token)
        return val if triggered else True

    def is_github_user(self, github_id: str, github_token: str) -> bool:
        def is_gh_user(gh_id: str, token: str) -> bool:
            res = requests.get(
                f"https://api.github.com/users/{gh_id}", headers=github_headers(token),
            )
            return res.status_code == 200

        triggered, val = self._safe_trigger(
            is_gh_user, gh_id=github_id, token=github_token
        )
        return val if triggered else True


def add_students(
    github_students: List[str] = typer.Argument(
        metavar="student_handles",
        default=...,
        help="list of student handles separated by white space",
    ),
    github_team: Optional[str] = typer.Option(
        None, "--team", help="invite to the specfic team under organization",
    ),
    yes: bool = opt_all_yes,
    dry: bool = opt_dry,
    github_token: Optional[str] = opt_github_token,
    github_organization: Optional[str] = opt_gh_org,
):
    """
    Invite students to join our Github organization
    """
    ensure_config_exists()

    def fallback(val, fallback_value):
        return val if val else fallback_value

    # Handle default value manually because we'll change our config after app starts up
    github_token: str = fallback(
        github_token, app_context.config.github.personal_access_token
    )
    github_organization: str = fallback(
        github_organization, app_context.config.github.organization
    )
    github_team: str = fallback(
        github_team, app_context.config.add_students.default_team_slug
    )

    safety = SafetyActor(dry=dry)
    safety.ensure_gh_token(github_token)

    # TODO: use logging lib to log messages
    spinner = Halo(stream=sys.stderr)
    if dry:
        spinner.info("Dry run")

    if not (
        yes or typer.confirm(f"Add students to {github_organization}/{github_team}?")
    ):
        raise typer.Abort()

    spinner.info("fetch existing team members from GitHub")
    team = Team(
        dry=dry,
        org=github_organization,
        team_slug=github_team,
        github_token=github_token,
    )
    num_member = len(team.members.keys())

    spinner.succeed(f" target team: {github_team} ({num_member} members) ")

    existed_members = set(team.members.keys())
    outside_users = list(set(github_students) - existed_members)
    spinner.info("Check valid Github users")
    invalid_handles = invalid_user_handles(
        outside_users, github_token=github_token, safety=safety, spinner=spinner
    )

    if len(invalid_handles) != 0:
        print("non-existed github user handles:")
        # control strings take space
        print_table(invalid_handles)
    non_member_valid_users = list(set(outside_users) - set(invalid_handles))

    # membership info
    spinner.info("Check Membership information")
    users_in_pending, users_to_invite = find_outside_members(
        non_member_valid_users, spinner=spinner, team=team, dry=dry
    )
    spinner.succeed("Check Membership information")

    if len(users_in_pending) > 0:
        print(f"Users already in pending state (total:{len(users_in_pending)}):")
        print_table(users_in_pending)

    print(f"Users to add (total:{len(users_to_invite)})")
    print_table(users_to_invite)
    print("-" * 30)

    spinner.info("start to invite users")
    success_user, failed_users = invite_user_to_team(
        team=team, users=users_to_invite, spinner=spinner
    )
    if len(failed_users) > 0:
        print("Users failed to add")
        print_table(failed_users)

    spinner.succeed("Add students successfully")


def invalid_user_handles(
    to_verify: List[str], github_token: str, safety: SafetyActor, spinner
) -> List[str]:
    """ return list of non-valid github user handles
        """
    invalid = []
    spinner.start()
    total = len(to_verify)
    for idx, user_name in enumerate(to_verify, start=1):
        text = f"{idx}/{total} Check valid GitHub username : {user_name}"
        if safety.is_github_user(user_name, github_token):
            spinner.succeed(text)
        else:
            invalid.append(user_name)
            spinner.fail(text)
    return invalid


def find_outside_members(
    user_handles: List[str], spinner, team: Team, dry: bool = False
) -> Tuple[List[str], List[str]]:
    """return two type of members,
        1. have invited but still in pending state
        2. not invited, not-organization users
        """
    pending_users, outside_users = [], []
    total = len(user_handles)
    for idx, username in enumerate(user_handles, start=1):
        skip = "" if not dry else "[skip]: "
        spinner.start(f"{skip}{idx}/{total}: {username}")
        res = team.get_memberships(username)
        if res.status_code == 200:
            state = res.json()["state"]
            if state == "pending":
                pending_users.append(username)
            elif state == "unknown":
                outside_users.append(username)
    return pending_users, outside_users


def invite_user_to_team(
    team: Team, users: List[str], spinner
) -> Tuple[List[str], List[str]]:
    """ invite user to team and return results in different name list
        """
    # make users unique
    users = list(set(users))
    success, failed, = [], []
    for user_name in users:
        res = team.add_user_to_team(user_name=user_name)
        if res.status_code == 200:
            success.append(user_name)
            spinner.succeed(f"add user: {user_name}")
        else:
            failed.append(user_name)
            spinner.fail(f"failed to add user: {user_name}")
    return success, failed



import sys
from typing import Dict

import httpx
import requests
from requests.models import Response

from invisible_hand.config import app_context
from invisible_hand.errors import ERR_CANNOT_FETCH_TEAM
from .github_scanner import (
    get_github_endpoint,
    get_github_endpoint_paged_list,
    github_headers,
)


def main():
    print("nothin to run")


class DrySuccessResponse:
    def __init__(self, code: int):
        self.status_code: int = code
        self.dict: Dict = {}

    def set_dict(self, dict: Dict):
        """ Setup the dict returned by self.json() method
        """
        self.dict = dict.copy()

    def json(self):
        return self.dict


class Team:
    def __init__(self, org, team_slug="", github_token="", dry=False):
        self.info_sucess = False
        self.id = ""
        self.org = org
        self.team_slug = team_slug
        self.members = dict()
        self.github_token = github_token

        # Support dry run, would not fire any request
        self.dry = dry

        if not dry:
            self._get_members()

        # async
        self.httpx_headers = httpx.Headers(
            {
                "User-Agent": "GitHubClassroomUtils/1.0",
                "Authorization": "token " + github_token,
                # needed for the check-suites request
                "Accept": "application/vnd.github.antiope-preview+json",
            }
        )
        self.async_client = httpx.AsyncClient(headers=self.httpx_headers)

    def add_user_to_team(self, user_name) -> Response:
        if self.dry:
            return DrySuccessResponse(200)
        else:
            res = requests.put(
                "https://api.github.com/teams/{}/memberships/{}".format(
                    self.id, user_name
                ),
                headers=github_headers(self.github_token),
            )
            return res

    async def add_team_repository_async(self, repo, permission="pull") -> Response:
        """If permission=pull, it is equivalent to make this team subscribe to this repo"""
        if self.dry:
            return DrySuccessResponse(204)
        else:
            api_param = {"team_id": self.id, "owner": self.org, "repo": repo}
            # print(f'param: {api_param}')
            return await self.async_client.put(
                "https://api.github.com/teams/{team_id}/repos/{owner}/{repo}".format(
                    **api_param
                )
            )

    def get_memberships(self, user_name: str) -> Response:
        if self.dry:
            response = DrySuccessResponse(200)
            # state id either "pending" or "unknown"
            response.set_dict({"state": "unknown"})
            return response
        else:
            return requests.get(
                "https://api.github.com/teams/{}/memberships/{}".format(
                    self.id, user_name
                ),
                headers=github_headers(self.github_token),
            )

    def _require_team_id(func):
        def decorator(self):
            if self.id == "":
                self._get_team_id()
            func(self)

        return decorator

    @_require_team_id
    def _get_members(self):
        """ fetch remote and build self.members
        """
        try:
            res = get_github_endpoint_paged_list(
                endpoint=f"teams/{self.id}/members",
                github_token=self.github_token,
                verbose=False,
            )
        except:
            raise ERR_CANNOT_FETCH_TEAM(org=self.org, team_slug=self.team_slug)
        for user in res:
            self.members[user["login"]] = user
        # print(res)

    def _get_team_id(self):
        try:
            result = get_github_endpoint(
                "orgs/{}/teams/{}".format(self.org, self.team_slug),
                app_context.config.github.personal_access_token,
            )
        except:
            raise ERR_CANNOT_FETCH_TEAM(org=self.org, team_slug=self.team_slug)
            sys.exit(1)
        self.id = result["id"]


if __name__ == "__main__":
    main()



import typer

opt_dry = typer.Option(False, "--dry", help="dry run, would not fire actual requests")

opt_all_yes = bool = typer.Option(
    False, "--yes", help="confirm to all", show_default=False
)


opt_github_token = typer.Option(
    None, "--token", help="personalaccess token used in Github API",
)

opt_gh_org = typer.Option(None, "--org", help="sepcify target github organization",)



from .base_config import Config
from .cmd import app
from .config_manager import ConfigManager
from .context import app_context

__all__ = [
    "app",
    "Config",
    "ConfigManager",
    "app_context",
]



# github_event_times.py
# Dan Wallach <dwallach@rice.edu>
# Available subject to the Apache 2.0 License
# https://www.apache.org/licenses/LICENSE-2.0

import re
import sys
from datetime import datetime, timedelta
from pathlib import Path
from pprint import pprint
from typing import List, NamedTuple, Optional, Tuple

import iso8601
import typer
from halo import Halo
from rich.console import Console
from rich.table import Table
from tabulate import tabulate

from invisible_hand.config import app_context
from ..ensures import ensure_config_exists, ensure_gh_token
from ..utils.github_entities import Team
from ..utils.github_scanner import (
    LOCAL_TIMEZONE,
    get_github_endpoint_paged_list,
    localtime_from_iso_datestr,
)
from .shared_options import opt_dry, opt_gh_org, opt_github_token


def is_deadline_passed(dl: datetime, submit: datetime) -> Tuple[bool, timedelta]:
    dl = dl.astimezone(submit.tzinfo)
    if dl < submit:
        return True, submit - dl
    else:
        return False, dl - submit


class commitInfo(NamedTuple):
    commit_hash: str  # 7 characters hash value
    pushed_time: str
    msg: str
    repo: str


class GitHubRepoInfo(NamedTuple):
    name: str
    commit_hash: str  # 7 characters hash value


# python3 github_event_times.py hw0-ianre657:cb75e99
class RepoPrinter:
    def __init__(
        self, repos: List[GitHubRepoInfo], title: str = "Repos",
    ):
        self.repos = repos
        self.title = title

    def print(self):
        table = Table(title=self.title)
        table.add_column("Name", justify="right", style="cyan")
        table.add_column("Commit Hash", justify="left")
        for r in self.repos:
            table.add_row(r.name, r.commit_hash)
        Console().print(table)


def event_times(
    input_file: str = typer.Argument(
        default=..., metavar="hw_path", help="file contains list of repo-hash"
    ),
    deadline: str = typer.Option(
        None,
        "--deadline",
        help="deadline format: 'yyyy-mm-dd' or any ISO8601 format string (timezone will be set to local timezone)",
    ),
    target_team: Optional[str] = typer.Option(
        default=None, help="specific team to operate on"
    ),
    dry: bool = opt_dry,
    token: Optional[str] = opt_github_token,
    org: Optional[str] = opt_gh_org,
):
    """
    Retrieve information about late submissions

    <repo-hash> : string in <repo>:<hash> format
            hw0-ianre657:cb75e99
    """
    ensure_config_exists()

    def fallback(val, fallback_value):
        return val if val else fallback_value

    # Handle default value manually because we'll change our config after app starts up
    deadline: str = fallback(deadline, app_context.config.event_times.deadline)
    token: str = fallback(token, app_context.config.github.personal_access_token)
    org: str = fallback(org, app_context.config.github.organization)

    global github_organization
    global github_token

    typer.echo(f"parse input file : {input_file}")
    try:
        parsed_repos = get_repo_infos(input_file)
    except FileNotFoundError as e:
        print(str(e))
        return
    RepoPrinter(parsed_repos).print()

    ensure_gh_token(token)
    spinner = Halo(stream=sys.stderr)

    github_organization = org
    github_token = token

    print(f"deadline: {deadline}")

    submit_deadline = iso8601.parse_date(deadline)
    submit_deadline = submit_deadline.replace(tzinfo=LOCAL_TIMEZONE)

    spinner.info(f"Deadline : {submit_deadline}")
    success_group = []
    fail_group = []
    spinner.start("Start to check late submissions")
    # sys.exit(0)
    # get team membershup info

    target_repos: List[GitHubRepoInfo] = list(parsed_repos)

    if target_team is not None:
        # Filter repos
        def get_handle(repo_name: str):
            """
                Deduce user handle from user name
                example: hw2-nctulaoda -> nctulaoda
            """
            return re.sub("hw[\d]+-", "", repo_name)

        target_team_members = set(
            Team(
                org=github_organization,
                team_slug=target_team,
                github_token=github_token,
                dry=dry,
            ).members.keys()
        )
        if not dry:
            target_repos = [
                r for r in target_repos if get_handle(r) in target_team_members
            ]

    for idx, repo in enumerate(target_repos, start=1):
        spinner.text = f"({idx}/{len(parsed_repos)}) Checking {repo.name}"
        if dry:
            continue
        result: Optional[commitInfo] = getCommitPushTime(
            org=github_organization, repo=repo.name, commit_hash=repo.commit_hash
        )
        if result:
            passed, delta = is_deadline_passed(
                submit_deadline, iso8601.parse_date(result.pushed_time)
            )
            if passed:
                fail_group.append(
                    {
                        "repo-name": result.repo,
                        "commit-hash": result.commit_hash,
                        "time-passed": delta,
                        "last-pushtime": result.pushed_time,
                    }
                )
            else:
                success_group.append((result, delta))

    spinner.succeed("Check finished")
    print("=" * 20, "REPORT", "=" * 20)
    print(f"Total submissions : {len(parsed_repos)}")
    print(f"late submissions: {len(fail_group)}")
    print(f"Submission Deadline: {submit_deadline}")
    print(tabulate(fail_group, headers="keys"))


def get_repo_infos(filename: str) -> List[GitHubRepoInfo]:
    st = Path(filename).read_text()
    infos = st.split()
    result = []
    for r in infos:
        repo, commit_hash = r.split(":")
        result.append(GitHubRepoInfo(repo, commit_hash))
    return result


def getCommitPushTime(org: str, repo: str, commit_hash: str) -> Optional[commitInfo]:
    """Find the push-time of given commit-hash
    """
    global github_token
    response = get_github_endpoint_paged_list(
        f"repos/{org}/{repo}/events", github_token, verbose=False
    )
    event_list = [x for x in response if x["type"] == "PushEvent"]
    for event in event_list:
        try:
            github_id = event["actor"]["login"]
            date = localtime_from_iso_datestr(event["created_at"])
            target_commits = [
                x for x in event["payload"]["commits"] if x["sha"][0:7] == commit_hash
            ]

            if len(target_commits) == 0:
                continue
            elif len(target_commits) >= 2:
                print("SHA Conflict for user: {}".format(github_id))
            else:
                target_commit = target_commits[0]

                commit_message = tex_escape(
                    target_commit["message"].splitlines()[0]
                )  # only the first line if multiline
                commit_hash = target_commit["sha"][0:7]

                return commitInfo(commit_hash, date, commit_message, repo)
        except KeyError:
            print("Error: malformed event!")
            pprint(event)
    return None


# https://stackoverflow.com/questions/16259923/how-can-i-escape-latex-special-characters-inside-django-templates


def tex_escape(text: str) -> str:
    """
    :param text: a plain text message
    :return: the message escaped to appear correctly in LaTeX
    """
    conv = {
        "&": r"\&",
        "%": r"\%",
        "$": r"\$",
        "#": r"\#",
        "_": r"\_",
        "{": r"\{",
        "}": r"\}",
        "~": r"\textasciitilde{}",
        "^": r"\^{}",
        "\\": r"\textbackslash{}",
        "<": r"\textless{}",
        ">": r"\textgreater{}",
    }
    regex = re.compile(
        "|".join(
            re.escape(str(key))
            for key in sorted(conv.keys(), key=lambda item: -len(item))
        )
    )
    return regex.sub(lambda match: conv[match.group()], text)



from .add_students import add_students
from .announce_grade import announce_grade
from .crawl_classroom import crawl_classroom
from .event_times import event_times
from .grant_read_access import grant_read_access
from .patch_project import patch_project

__all__ = [
    "add_students",
    "grant_read_access",
    "announce_grade",
    "crawl_classroom",
    "event_times",
    "patch_project",
]



from typing import Dict, Optional

from tomlkit import comment, document, dumps, nl, table

from .base_config import Config


class _ERR_NON_VAILD_DEFAULT_CFG(Exception):
    """ Internal use,
    make sure that our default config is always valid
    """

    def __init__(self):
        super().__init__(self)


def add_commented_section(
    docu: document, section_name: str, dic: Dict, section_comment: Optional[str] = None,
):
    """ Add a section with comment
    """

    def build_table(header=None):
        try:
            t = table()
            if header:
                t.add(comment(header))

            for k, v in dic.items():
                if isinstance(v, Dict):
                    if v.get("comment"):
                        t.add(comment(v["comment"]))
                    elif v.get("comments"):
                        for c in v["comments"]:
                            t.add(comment(c))
                    else:
                        raise _ERR_NON_VAILD_DEFAULT_CFG
                    t.add(k, v["value"])
                    t.add(nl())
                else:
                    t.add(k, v)
        except:
            raise _ERR_NON_VAILD_DEFAULT_CFG
        return t

    docu.add(section_name, build_table(header=section_comment))


def export_default_config() -> str:
    """ Generate the default config file
    """

    def add_big_title(docu, title: str, width: int = 50):
        def side(width):
            return "|" + " " * (width - 2) + "|"

        lines = [
            "=" * width,
            side(width),
            f"|{title.center(width-2)}|",
            side(width),
            "=" * width,
        ]
        for _ in range(3):
            docu.add(nl())
        for line in lines:
            docu.add(comment(line))

    default_cfg = Config.get_default()
    docu = document()
    docu.add(comment("Config file for invisible-hand"))
    docu.add(comment("each section define the corresponding default config for scripts"))
    add_big_title(docu, "Env Configuration")

    add_commented_section(
        docu=docu,
        section_name="github",
        dic={
            "organization": {
                "comment": "Your GitHub organization (https://github.com/Organization/Repository/...)",
                "value": default_cfg.github.organization,
            },
            "personal_access_token": {
                "comments": [
                    "Your GitHub API token here, inside the quotation marks",
                    "https://github.com/blog/1509-personal-api-tokens",
                ],
                "value": default_cfg.github.personal_access_token,
            },
        },
    )

    add_commented_section(
        docu=docu,
        section_name="google_spreadsheet",
        section_comment="You need to config this section before using `hand annouce-grade`",
        dic={"spreadsheet_url": default_cfg.google_spreadsheet.spreadsheet_url,},
    )

    add_big_title(docu, "Scripts Configuration")

    add_commented_section(
        docu=docu,
        section_name="add_students",
        dic={"default_team_slug": default_cfg.add_students.default_team_slug},
    )

    add_commented_section(
        docu=docu,
        section_name="grant_read_access",
        dic={"reader_team_slug": default_cfg.grant_read_access.reader_team_slug,},
    )

    add_commented_section(
        docu=docu,
        section_name="crawl_classroom",
        dic={
            "login": {
                "comment": "Your login id in Github Classroom",
                "value": default_cfg.crawl_classroom.login,
            },
            "classroom_id": {
                "comment": "id field of your classroom RESTful page URL.",
                "value": default_cfg.crawl_classroom.classroom_id,
            },
        },
    )

    add_commented_section(
        docu=docu,
        section_name="event_times",
        dic={
            "deadline": {
                "comments": [
                    "Deadline for homework, in ISO8601 compatible format",
                    "For example `2019-11-12 23:59:59` (the timezone is set to your local timezone as default)",
                    "Or simply `2019-11-12`",
                ],
                "value": default_cfg.event_times.deadline,
            }
        },
    )

    add_commented_section(
        docu=docu,
        section_name="announce_grade",
        dic={
            "feedback_source_repo": {
                "comment": "You're not supposed to change this variable",
                "value": default_cfg.announce_grade.feedback_source_repo,
            }
        },
    )

    return dumps(docu)



from typing import Optional

from pydantic import BaseModel

from .config_manager import Config, ConfigManager


class AppContext(BaseModel):
    config_manager: ConfigManager = ConfigManager()
    config: Optional[Config] = None

    class Config:
        arbitrary_types_allowed = True


app_context = AppContext()



from pydantic import BaseModel


class Config(BaseModel):
    """
        The Config class
    """

    class Github(BaseModel):
        personal_access_token: str
        organization: str

    class Google(BaseModel):
        spreadsheet_url: str

    class CrawlClassroom(BaseModel):
        login: str
        classroom_id: str

    class GrantReadAccess(BaseModel):
        reader_team_slug: str

    class AddStudens(BaseModel):
        default_team_slug: str

    class EventTimes(BaseModel):
        deadline: str

    class AnnounceGrade(BaseModel):
        feedback_source_repo: str

    github: Github
    google_spreadsheet: Google
    crawl_classroom: CrawlClassroom
    grant_read_access: GrantReadAccess
    add_students: AddStudens
    event_times: EventTimes
    announce_grade: AnnounceGrade

    @staticmethod
    def get_default():
        return Config(
            github=Config.Github(
                personal_access_token="your-github-personal-access-token",
                organization="compiler-s20",
            ),
            google_spreadsheet=Config.Google(spreadsheet_url="your-spreadsheet-id"),
            crawl_classroom=Config.CrawlClassroom(
                login="your-github-id", classroom_id="classroom-id-in-url"
            ),
            grant_read_access=Config.GrantReadAccess(
                reader_team_slug="2020-teaching-team"
            ),
            add_students=Config.AddStudens(default_team_slug="2020-students"),
            event_times=Config.EventTimes(deadline="2019-11-12 23:59:59"),
            announce_grade=Config.AnnounceGrade(feedback_source_repo="Hw-manager"),
        )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['invisible_hand',
 'invisible_hand.config',
 'invisible_hand.core',
 'invisible_hand.scripts',
 'invisible_hand.scripts.shared_options',
 'invisible_hand.utils']

package_data = \
{'': ['*']}

install_requires = \
['attrs>=20.3.0,<21.0.0',
 'beautifulsoup4>=4.9.1,<5.0.0',
 'click>=7.1.2,<8.0.0',
 'colorama>=0.4.3,<0.5.0',
 'gitpython>=3.1.7,<4.0.0',
 'google-api-python-client>=1.10.0,<2.0.0',
 'google-auth-httplib2>=0.0.4,<0.0.5',
 'google-auth-oauthlib>=0.4.1,<0.5.0',
 'halo>=0.0.30,<0.0.31',
 'httpx>=0.14.1,<0.15.0',
 'iso8601>=0.1.12,<0.2.0',
 'lxml>=4.5.2,<5.0.0',
 'oauth2client>=4.1.3,<5.0.0',
 'prompt-toolkit>=3.0.6,<4.0.0',
 'pydantic>=1.7,<2.0',
 'pygsheets>=2.0.3,<3.0.0',
 'requests>=2.24.0,<3.0.0',
 'rich>=9.2.0,<10.0.0',
 'selenium>=3.141.0,<4.0.0',
 'tabulate>=0.8.7,<0.9.0',
 'tomlkit>=0.7.0,<0.8.0',
 'tqdm>=4.48.2,<5.0.0',
 'trio>=0.16.0,<0.17.0',
 'typer[all]>=0.3.2,<0.4.0',
 'xlsxwriter>=1.3.3,<2.0.0']

entry_points = \
{'console_scripts': ['hand = invisible_hand.cli:app']}

setup_kwargs = {
    'name': 'invisible-hand',
    'version': '0.3.0',
    'description': 'Automate your workflow with github classroom',
    'long_description': '# Invisible Hand 👊\n\nInvisible Hand helps you manage your classroom under Github organization.<br>\nIt utilizes `Google Sheets` , `GitHub Classroom` and `GitHub` seamlessly.\n\n## Installation\n\n#### 1. Install this tool via pip\n\n `pip install invisible-hand`\n\n Use `hand -h` to test if it is installed\n\n```\n╰─❯ hand -h\nUsage: hand [OPTIONS] COMMAND [ARGS]...\n\n  Manage your classroom under Github organization.\n\nOptions:\n  --custom-base TEXT    Use custom base folder for configs\n  --install-completion  Install completion for the current shell.\n  --show-completion     Show completion for the current shell, to copy it or\n                        customize the installation.\n\n  -h, --help            Show this message and exit.\n\nCommands:\n  add-students       Invite students to join our Github organization\n  announce-grade     Announce student grades to each hw repo\n  config             Config File utilities\n  crawl-classroom    Get student\'s submissions from Github Classroom\n  event-times        Retrieve information about late submissions <repo-\n                     hash>...\n\n  grant-read-access  Make TAs being able to read all homework repos...\n  patch-project      Patch to student homeworks\n```\n\n#### 2. Install chromedriver\n\n* Mac : `brew cask install chromedriver`\n* Ubuntu : `apt install chromium-chromedriver`\n\n## Getting Started\n\n**1. Create config file**\n\nThe main config file is located in `$HOME/.invisible-hand/config.toml`. <br>\nRun `hand config create` to create a template and use `hand config edit` to open it with an editor.\n\n**2. `(Not required)` Getting your `client_secret.json` for accessing GoogleSheets**\n\nTo run the command:[ `Announce Grade` ](#announce-grade), follow the steps below to setup your environment first\n\n1. Follow [Authorization for pygsheets][pygsheet-auth] to get your credential file (`client_secret.json`)\n2. Locate your secret file and use `hand config copy-client-secret YOUR-SECRET-FILE` to copy it into the cache.\n\n    This command would automatically rename your secret file to (`$HOME/.invisible-hand/client_secret.json`)\n\n[pygsheet-auth]: https://pygsheets.readthedocs.io/en/stable/authorization.html\n\n## Quick references\n\n(This tool supports auto-completion, use `hand --install-completion` to add completion support to your shell)\n\n+ `hand`: The root command, use `-h` to see more information\n  + `config`: subcommand to handle config files, use `-h` to see more information\n    + `create`: Create a config file from template to `$HOME/.invisible-hand/config.toml`\n    + `check`: Check if the format your config file is valid\n    + `path`: Show the root path to your config folder, default to `$HOME/.invisble-hand`\n    + `edit`: Open your config file with your default editor. (use `--editor <your editor>` to open your another one)\n    + `copy-client-secret`: Copy Google\'s `client_secret.json` to the cache folder.\n      + use `hand config copy-client-secret -h` to see more information\n  + [`add-students 🧑\u200d💼student1 🦸\u200d♂️student2 ...`][add-students]: Invite users into your GitHub organization\n    + `🧑\u200d💼student`- GitHub handle of the student\n  + `grant-read-access 📝 hw-title`:  Grant TA\'s group read access to all repo with such prefix.\n    + `📝 hw-title`- title prefix of the homework (e.g. `hw3`)\n  + [`patch-project 📝 hw-title  🚧 patch-branch`][patch-project]: Patch project to Students homeworks\n  + [`crawl-classroom 📝 hw-title 📦 submission-file`][crawl-classroom]: Crawl submission status of students on GitHub Classroom\n    + `📦 submission-file`- file to store submission output\n  + [`event-times 📦 submission-file`][event-times]: Check if there\'s late submissions\n  + [`announce-grade 📝 hw-title`][announce-grade]: Announce grade to students by opening issues in their repos\n\n\n---\n\n## Add Students\n\n#### FAQ\n\n* Some students report that they didn\'t get the invitation email.\n\n    Invite student into your organization from their email. This should be Github\'s issue.\n\n    > about 2 of 80 students got this issue from our previous experience.\n\n\n\n## Patch Project\n\nPatch to student homework repositories.\n\n**Workflow**\n\nTake homework : **`hw3`**(the title of your homework in github classroom) for example: <br>\nThe repository: **`tmpl-hw3`** would be your template for initializing homeworks.\n\n#### 1. Create another repo with name **`tmpl-hw3-revise`** to update your template\n\n#### 2. Inside **`tmpl-hw3-revise`**\n\n1. Create a revision branch **`2-revise-for-node-definition`** (whatever you like)\n\n<p align="center">\n<img src="./imgs/patch-branches.png" alt="name of the branch" width="640">\n</p>\n\n2. Create an issuse with the identiacal title **`2-revise-for-node-definition` )**\n   1.  the content of that issue would become the content of your PR message.\n\n<p align="center">\n<img src="./imgs/patch-template-issue.png" alt=" template issue" width="640">\n</p>\n\n#### 3. Open GitHub Classroom\n\nSelect your assignment (**`hw3`**) and disable `assignment invitation URL` of **`hw3`**\n\n<p align="center">\n<img src="./imgs/patch-diable-hw.png" alt="disable invitation url" width="640">\n</p>\n\n#### 4. Create an PR to your template repo (**`hw3`**)\n\n``` sh\nhand patch-project hw3 --only-repo="tmpl-hw3" 2-revise-for-node-definition\n```\n\n#### 5. Accept the PR in your template repository (**`tmpl-hw3`**)\n\nAfter that, enable the `assignment invitation URL` of `hw3` in GitHub Classroom.<br>\nNow you have succcessfully updated your template repo.\n\n#### 6. Create PRs to students template repositories ( `hw3-<their github id>` )\n\nPatch to every repository that uses **hw3** as the prefix under your GitHub organization.\n\n``` sh\nhand patch-project hw3 2-revise-for-node-definition\n```\n\n#### 7. Merge the revision branch **`2-revise-for-node-definition`** inside `tmpl-hw3`\n\n\nAfter this step, all documents are updated\n\n#### 8. Reactivate the invitation URL\n\n<p align="center">\n<img src="./imgs/patch-enable-hw.png" alt="reactivate invitation url" width="640">\n</p>\n\n\n## Crawl Classroom\n\nCrawling homework submission data from Github Classroom\n\nThis is a web crawler for Github Classroom, which is the input of [ `Event Times` ][event-times]\n\n### Config File\n\n* **config.toml**:\n  + `[crawl_classroom]:login` : your login id in Github Classroom\n  + `[crawl_classroom]:classroom_id` : the id field of your classroom RESTful page URL. (see the image below)\n\n<p align="center">\n <img src="./imgs/clsrm_id.png" alt="id field in the url of github classroom" width="640">\n</p>\n\n\n\n### FAQ\n\n* ChromeDriver\n\n  ```\n  selenium.common.exceptions.SessionNotCreatedException: Message: session not created:\n  This version of ChromeDriver only supports Chrome version 79\n  ```\n\n    upgrade your chromedriver via `brew cask upgrade chromedriver`\n  * All students not submitted\n    + Remember to set deadline of hw on the GitHub classroom (note that deadline can only be set at a future time)\n\n<p align="center">\n<strong>Demo</strong><br>\n<img src="./demos/github_classroom_craw.gif" alt="github_classroom_craw" width="640">\n</p>\n\n\n\n\n## Event Times\n\nRetrieve information about late submissions\n\n**What it actually does**\n\nCompare the last publish-time of specific git commit in each repository and print out which passed the deadline.\n\n\n**Example**\n\n``` sh\nhand event-times  --target-team="2020-inservice-students" --deadline="2019-11-12 23:59:59"  hw1-handin-0408.txt\n```\n\n\n\n## Announce Grade\n\nPublish feedbacks by creating Issue to student\'s homework repo.\n\nThis sciprt requires you to\n  + Use a `GoogleSpreadSheet` to record every student\'s grade inside a strictly named tab.\n  + Strictly structured Github repository.\n\n\n**Explanation**\n\nIn our scheme, each student would get a git repository for every homework project.<br>\n\nTake homework `hw3` with two students `Anna` and `Bella` for example.<br>\nWe expect there would be 2 repos under our GitHub organization, which are `hw3-Anna` and `hw3-Bella`.<br>\n\nAnnoucing the grade for `hw3` actually means to open 1 issue to `hw3-anna` and `hw3-bella` respectively.\n\nTo make sure the script correctly functions:\n+ students grade should be recorded in the SpreadSheet with a tab called `hw3`.\n+ You should keep a `Markdown` file for each student using their `student-id` as the filename to describe their homework result.\n  + e.g. (if `Anna`\'s sutdent-id is `0856039`, then `0856039.md` should record `Anna`\'s overall feedbacks).\n  + **Behold: `student-id` should be unique.**\n\n**Format and Structure for those files**\n\n+ 0856039.md (`Anna`)\n\n  ```markdown\n  Hi, @Anna\n\n  ## Information\n  + Grade: ${grade}\n    + testcase : ${grade_testcase}/100 pts\n    + report : ${grade_report}/5 pts\n\n  + Grader: @{grader}\n  + Feedback:\n      Good Job\n\n  ## Note\n  If you got any questions about your grade\n  please make a comment stating your arguments below and tag the grader in the comment.\n  ```\n\n  This markdown file contains python template strings(`${grade}`),<br>\n  These strings are the column names inside your SpreadSheet tab `hw3`.\n\n+ Structure of the GitHub Repo:\n  ``` markdown\n    . Hw-manager # root of your git repo (the name is configured in `$HOME/.invisible-hand/config.toml`)\n    ├── hw3\n    │\xa0\xa0 └── reports\n    │\xa0\xa0     ├── 0411276.md\n    │\xa0\xa0     ├── 0856039.md (**Anna**)\n    └── hw4 # other homework dir\n  ```\n\n+ Structure of the Google SpreadSheet\n  | student_id | grade | grade_report | grade_testcase | grader |\n  | :--------: | :---: | :----------: | :------------: | :----: |\n  |  0856039   |  93   |      5       |       87       |  @TA1  |\n  |  0411276   |  80   |      5       |       75       |  @TA2  |\n\n\n### Config file\n\n* **config.toml**\n  + `[github]:personal_access_token`\n  + `[github]:organization`\n  + `[announce_grade]:feedback_source_repo` (e.x.: Hw-manager)\n  + `[google_spreadsheet]:spreadsheet_id`\n* **client_secret.json** (follow [here](https://pygsheets.readthedocs.io/en/stable/authorization.html) to download your oauth2 secret file and renamed it to **client_secret.json**)\n\n#### Instructions to follow\n\n1. Edit config files properly.\n2. Create feedbacks for students in your `feedback_source_repo`\n   1. **Make sure you commit and push the feedbacks to `master` branch**\n   2. Fill out the SpreadSheet\n3. Use this script\n   1. For testing, it\'s recommended to use the `--only-id` option\n\n       ```sh\n       hand announce-grade <hw_title> [--only-id 👨\u200d🎓student-id]\n       ```\n\n\n[event-times]: #event-times\n[add-students]: #add-students\n[patch-project]: #patch-project\n[crawl-classroom]: #crawl-classroom\n[announce-grade]: #announce-grade',
    'author': 'Ian Chen',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from typing import Optional

import typer

from invisible_hand.ensures import ensure_client_secret_json_exists


def dev(
    org: Optional[str] = typer.Option(None, help="Last name of person to greet."),
    age: str = typer.Option("", help="Some help message"),
):
    ensure_client_secret_json_exists()
    print("nice")



import textwrap
from pprint import pformat
from typing import Dict, List, Optional

from .urls import url_for_issue


def format_paragraph(para: List[str]) -> str:
    result = ["\n\n"]
    for line in para:
        result += f"{line}\n"
    return textwrap.indent("".join(result), prefix=" " * 4)


class ERR_GOOGLE_CLIENT_SECRET_NOT_EXISTED(Exception):
    def __str__(self):
        exps = [
            "Didn't detect a `client_secret.json` file inside your cache folder",
            "If you already have it, use",
            "   `hand config copy-client <path-to-your-client-secret-file>",
            "to copy it into your cache folder",
            "",
            "If you don't know how to get this file, follow",
            "   https://pygsheets.readthedocs.io/en/stable/authorization.html",
            "to get your client_secret file, and follow the step above to copy it to the cache folder",
        ]
        return format_paragraph(exps)


class ERR_CHROME_DRIVER_NOT_INSTALLED(Exception):
    def __str__(self):
        exps = [
            "`chromedriver` not installed",
            "You need to install it before going further.",
            "If your'e using ubuntu, run:",
            "   `apt install chromium-chromedriver`",
            "For Mac users, run:",
            "   `brew cask install chromedriver`",
        ]
        return format_paragraph(exps)


class ERR_CONFIG_NOT_EXISTS(Exception):
    def __str__(self):
        return """
        Config not exists
        Use `hand config create` to create a new config file
        """


class ERR_UNIQUE_STUDENT_ID(Exception):
    def __init__(self, explanation: str, instances: List[Dict]):
        super().__init__(self)
        self.explanation: str = explanation
        self.instances: List[Dict] = instances

    def __str__(self):
        words = [self.explanation, pformat(self.instances)]
        return "\n".join(words)


class ERR_REQUIRE_NO_SPACE(Exception):
    def __init__(self, explanation: str, instances: List[str]):
        super().__init__(self)
        self.explanation: str = explanation
        self.instances: List[str] = instances

    def __str__(self):
        words = [self.explanation, pformat(self.instances)]
        return "\n".join(words)


class ERR_GIT_CREDENTIAL_HELPER_NOT_FOUND(Exception):
    def __str__(self):
        issue_git_cred = url_for_issue(issue_num=8)
        exps = [
            "Not detecting any credential helper inside git, use:",
            "   git config --global credential.helper cache",
            "before going any further.",
            "Or see:",
            f"  {issue_git_cred}",
            "for more information",
        ]
        return format_paragraph(exps)


class ERR_INVALID_GITHUB_TOKEN(Exception):
    def __init__(self, token: Optional[str] = None):
        self.token = token
        super().__init__(self)

    def __str__(self) -> str:
        exps = [
            "Invalid Github token" + (f": {self.token}." if self.token else ""),
            "Please update `github_config.ini`.",
        ]
        return format_paragraph(exps)


class ERR_CANNOT_FETCH_TEAM(Exception):
    def __init__(self, org: str, team_slug: str):
        self.org = org
        self.team = team_slug
        super().__init__(self, f"cannot fetch team: {self.org}/{self.team}")

    def __str__(self) -> str:
        exps = [
            "Cannot fetch team id",
            "Team info :",
            f"     org :{ self.org}",
            f"     team_slig :{ self.team}",
        ]
        return format_paragraph(exps)



import os
from pathlib import Path
from shutil import copy2, which

import typer
from pydantic import ValidationError

from .config_manager import ConfigManager
from .context import app_context
from .words import STR_CFG_IS_VALID, STR_CFG_NOT_EXISTS

manager: ConfigManager

app = typer.Typer(help="Config File utilities")


@app.callback()
def setup_manager():
    global manager
    manager = app_context.config_manager


@app.command()
def create():
    """Create config file"""
    global manager

    config_path = manager.config_path
    if not manager.config_path.exists():
        manager.create_default()
        typer.echo(f"Config file created: {config_path}")
        raise typer.Exit(0)

    override = typer.confirm("Config file exists, override it?")
    if not override:
        raise typer.Abort()
    manager.create_default()
    typer.echo(f"Config file created: {config_path}")
    raise typer.Exit(0)


@app.command("copy-client-secret")
def copy_client_secret(
    src_path: Path = typer.Argument(
        default=..., metavar="📁file", help="Path to your client_secret file"
    ),
):
    """Copy client_secret.json to cache folder"""
    global manager
    dst_path = manager.google_client_secret_path
    if not src_path.exists():
        typer.echo(f"File not exists: {src_path}")
        typer.Abort()
    copy2(src_path, dst_path)
    typer.echo(f"Copy file:{src_path} to {dst_path}")


@app.command()
def check():
    """Check if your config file is valid & usable"""
    global manager

    if not manager.config_path.exists():
        typer.echo(STR_CFG_NOT_EXISTS)
        raise typer.Exit(1)

    try:
        _ = manager.read_config()
    except ValidationError as e:
        typer.echo(e)
        typer.echo(f"please check your config file: {manager.config_path}")
        raise typer.Abort()

    typer.echo(STR_CFG_IS_VALID)
    raise typer.Exit(0)


@app.command()
def edit(editor: str = os.getenv("EDITOR", default="code")):
    """Open config file"""
    global manager
    if which(editor) is not None:
        os.system(f"{editor} {manager.config_path}")
    else:
        typer.echo("Please specify a valid editor to open the file")
        raise typer.Abort()


@app.command()
def remove(
    yes: bool = typer.Option(False, "--yes", help="confirm to remove", show_default=False)
):
    """Remove related data completely"""
    global manager
    base_path: Path = manager.get_base_path().resolve()
    confirm = yes or typer.confirm(f"Remove folder : {base_path}?")
    if not confirm:
        typer.echo("Not Removing")
        raise typer.Abort()
    manager.remove()
    typer.echo(f"Folder Removed: {base_path}")


@app.command()
def path():
    """Get the folder for configs"""
    global manager
    base_path: Path = manager.get_base_path().resolve()
    typer.echo(base_path)



import os
import shutil
import subprocess as sp
import sys
import tempfile
from datetime import datetime
from pathlib import Path
from typing import Dict, Optional

import requests
import typer
from git import Repo
from git.objects.commit import Commit
from halo import Halo

from invisible_hand.config import app_context
from invisible_hand.ensures import (
    ensure_config_exists,
    ensure_gh_token,
    ensure_git_cached,
)
from ..utils.github_scanner import (
    get_github_endpoint_paged_list,
    github_headers,
    query_matching_repos,
)
from .shared_options import opt_all_yes, opt_gh_org, opt_github_token

# The Pull request we made would fetch the first issue which has the same title
# as patch_branch to be the content


def patch_project(
    hw_prefix: str = typer.Argument(default=..., help="prefix of the homework title"),
    patch_branch: str = typer.Argument(
        default=..., help="source bracnch to patch to the main branch"
    ),
    source_repo: Optional[str] = typer.Option(
        default=None, help="default to tmpl-{hw-prefix}-revise"
    ),
    only_repo: Optional[str] = typer.Option(default=None, help="only repo to patch"),
    dry: bool = typer.Option(
        False, "--dry", help="dry run, do not publish result to the remote"
    ),
    yes: bool = opt_all_yes,
    token: str = opt_github_token,
    org: str = opt_gh_org,
):
    """Patch to student homeworks"""
    ensure_config_exists()

    def fallback(val, fallback_value):
        return val if val else fallback_value

    # Handle default value manually because we'll change our config after app starts up
    token: str = fallback(token, app_context.config.github.personal_access_token)
    org: str = fallback(org, app_context.config.github.organization)

    ensure_git_cached()
    ensure_gh_token(token)

    if not (
        yes or typer.confirm(f"Patch homework:{hw_prefix} from branch:{patch_branch}?")
    ):
        raise typer.Abort()
    # init
    spinner = Halo(stream=sys.stderr)

    if not source_repo:
        source_repo = f"tmpl-{hw_prefix}-revise"

    # Check if repo already contains the patched branch. Skip if so.
    #  api : https://developer.github.com/v3/git/refs/#get-a-reference
    res = requests.get(
        f"https://api.github.com/repos/{org}/{source_repo}/git/refs/heads/{patch_branch}",
        headers=github_headers(token),
    )
    if res.status_code != 200:  # this branch not exists on the remote
        spinner.fail(
            f"branch : `{patch_branch}` doesn't exist on repo:{org}/{source_repo} "
        )
        return

    cur = Path(".")
    for d in cur.glob("patch-*"):
        shutil.rmtree(d)
    spinner.info("delete dated folder")

    spinner.start(f"Fetch issue template {patch_branch} from {source_repo}")
    # Fetch patch template on the source repo
    issues = get_github_endpoint_paged_list(
        endpoint=f"repos/{org}/{source_repo}/issues", github_token=token, verbose=False
    )

    def find_target_issue() -> Optional[Dict]:
        for issue in issues:
            if issue["title"].strip() == patch_branch.strip():
                return issue
        return None

    target_issue = find_target_issue()
    if not target_issue:
        raise Exception(f"cannot found issue tmpl `{patch_branch}` on `{source_repo}`")
    issue_tmpl_body = target_issue["body"]
    spinner.succeed()

    root_folder = Path(
        tempfile.mkdtemp(
            prefix="patch-{}-{}-".format(
                patch_branch, datetime.now().strftime("%b%d%H%M%S")
            ),
            dir=".",
        )
    )

    spinner.succeed(f"Create tmp folder: {root_folder}")
    spinner.info(f"Fetch source repo {source_repo} from Github")
    src_repo_path = root_folder / "source_repo"
    sp.run(
        [
            "git",
            "clone",
            f"https://github.com/{org}/{source_repo}.git",
            src_repo_path.name,
        ],
        cwd=root_folder,
    )

    src_repo = Repo(src_repo_path)
    sp.run(
        ["git", "checkout", "--track", f"origin/{patch_branch}"],
        cwd=src_repo_path,
        stdout=sp.DEVNULL,
        stderr=sp.DEVNULL,
    )
    spinner.succeed()

    # Pasting changed files into students repo
    src_repo_git = src_repo.git
    src_repo_git.checkout(patch_branch)
    changed_files, renamed_files = get_changed_files(
        master_commit=src_repo.heads["master"].commit,
        patch_commit=src_repo.heads[patch_branch].commit,
    )

    spinner.start("Fetch information for homework repo")
    spinner.succeed()
    if only_repo is not None:
        repos = [
            re
            for re in query_matching_repos(
                org, github_repo_prefix=only_repo, github_token=token, verbose=False
            )
            if re["name"] == only_repo
        ]
        repo = next(iter(repos), None)
        if repo:
            spinner.info("Only patch to repo : " + repo["name"])
        repos = [repo]
    else:
        repos = query_matching_repos(
            org, github_repo_prefix=hw_prefix, github_token=token, verbose=False
        )
    spinner.succeed()

    # Patch to student repos
    student_path = root_folder / "student_repos"
    student_path.mkdir()
    for repo_idx, r in enumerate(repos, start=1):
        pre_prompt_str = f"({repo_idx}/{len(repos)}) " + r["name"]
        spinner.start()

        # Check if repo already contains the patched branch. Skip if so.
        #  api : https://developer.github.com/v3/git/refs/#get-a-reference
        res = requests.get(
            f"https://api.github.com/repos/{org}/{r['name']}/git/refs/heads/{patch_branch}",
            headers=github_headers(token),
        )
        if res.status_code == 200:  # this branch exists in the remote
            spinner.text = pre_prompt_str + " Skip " + " already patched"

            spinner.succeed()
            continue

        spinner.text = pre_prompt_str + "cloning repo"

        sp.run(
            ["git", "clone", "--depth=1", r["html_url"]],
            cwd=student_path,
            stdout=sp.DEVNULL,
            stderr=sp.DEVNULL,
        )

        hw_repo_name = r["html_url"].rsplit("/")[-1]

        # open a new branch & checkout to that branch
        sp.run(
            ["git", "checkout", "-b", patch_branch],
            cwd=student_path / hw_repo_name,
            stdout=sp.DEVNULL,
            stderr=sp.DEVNULL,
        )

        # copy file to student repo
        for f in changed_files.keys():
            (student_path / hw_repo_name / f).parent.mkdir(parents=True, exist_ok=True)
            shutil.copyfile(src=src_repo_path / f, dst=student_path / hw_repo_name / f)
        for f in renamed_files.keys():
            os.remove(student_path / hw_repo_name / f)

        # changed_files = get_changed_files(
        #     master_commit = src_repo.heads['master'].commit,
        #     patch_commit = src_repo.heads[patch_branch].commit
        # )
        # push (publish) that branch to student repo
        sp.run(
            ["git", "add", "."],
            cwd=student_path / hw_repo_name,
            stdout=sp.DEVNULL,
            stderr=sp.DEVNULL,
        )

        # Pass if no changed
        student_repo = Repo(student_path / hw_repo_name)
        if len(student_repo.index.diff("HEAD")) == 0:
            spinner.text = pre_prompt_str + " Passed " + " no changes in repo"

            spinner.succeed()
            continue

        sp.run(
            ["git", "commit", "-m", f":construction_worker: Patch: {patch_branch}"],
            cwd=student_path / hw_repo_name,
            stdout=sp.DEVNULL,
            stderr=sp.DEVNULL,
        )

        spinner.text = pre_prompt_str + " publish patch to remote..."
        if dry:
            spinner.succeed(pre_prompt_str + " Patched ")
            continue
        res = sp.run(
            ["git", "push", "-u", "origin", patch_branch],
            cwd=student_path / hw_repo_name,
            stdout=sp.DEVNULL,
            stderr=sp.DEVNULL,
        )
        if res.returncode != 0:
            spinner.fail(
                pre_prompt_str
                + "  Failed  "
                + f"Cannot push branch {patch_branch} to_origin"
            )
            continue

        # open an pull-request on students repo
        # student_repo/patch-branch  -> student_repo/master
        body = {
            "title": f"[PATCH] {patch_branch}",
            "body": issue_tmpl_body,
            "head": patch_branch,
            "base": "master",
        }
        res = requests.post(
            f"https://api.github.com/repos/{org}/{r['name']}/pulls",
            headers=github_headers(token),
            json=body,
        )
        if res.status_code == 201:
            spinner.succeed(f"{pre_prompt_str} Patched ")
        else:
            spinner.fail(
                pre_prompt_str
                + "  Failed  "
                + f"Cannot create PR {patch_branch} to origin/master"
            )
            try:
                print(res.json()["errors"][0]["message"])
            except:
                pass
            continue

    # TODO : print summary after patch
    #       how many success, skiped, failed


def get_changed_files(master_commit: Commit, patch_commit: Commit):
    # TODO
    #  git change type support:
    #  A: addition of a file
    #  C: copy of a file into a new one
    #  D: deletion of a file
    # done  M: modification of the contents or mode of a file
    #  R: renaming of a file
    #  T: change in the type of the file
    #  U: file is unmerged (you must complete the merge before it can be committed)
    #  X: "unknown" change type (most probably a bug, please report it)
    changed_files = {}
    renamed_files = {}
    for x in master_commit.diff(patch_commit):
        # Change type of x is not 'new'
        if x.a_blob != None and x.a_blob.path not in changed_files.keys():
            if x.change_type == "R":
                # file have been renamed, the dest file is include in the changed files
                # so we need to delete this file from dest repo

                # print(f'a remove(rename) :{x.a_blob.path}, type: {x.change_type}')
                renamed_files[x.a_blob.path] = {"type": x.change_type}
            else:
                # print(f'a change :{x.a_blob.path}, type: {x.change_type}')
                changed_files[x.a_blob.path] = {"type": x.change_type}

        # Change type of x is not 'delete'
        if x.b_blob is not None and x.b_blob.path not in changed_files.keys():
            # print(f'b change :{x.b_blob.path}, type: {x.change_type}')
            changed_files[x.b_blob.path] = {"type": x.change_type}

    return changed_files, renamed_files


if __name__ == "__main__":
    patch_project()



import shutil
from pathlib import Path
from typing import Union

import attr
from tomlkit import parse

from .base_config import Config
from .export_toml import export_default_config
from .words import STR_CFG_NOT_EXISTS

DEFAULT_BASE_FOLDER = Path.home() / ".invisible-hand"
CFG_FILENAME = "config.toml"
CLIENT_SECRET_FILENAME = "client_secret.json"


class CONFIG_NOT_EXISTS(Exception):
    def __init__(self, path):
        self.path = path
        super().__init__(self)

    def __str__(self):
        return f"{STR_CFG_NOT_EXISTS}: {self.path}"


@attr.s(auto_attribs=True)
class ConfigManager:
    base_folder: Path = attr.ib(default=DEFAULT_BASE_FOLDER, converter=Path)
    config_path: Path = attr.ib(init=False)
    google_client_secret_path: Path = attr.ib(init=False)

    def __attrs_post_init__(self):
        self.change_base_folder(self.base_folder)

    def change_base_folder(self, new_base: Union[Path, str]):
        self.base_folder = Path(new_base)
        self.config_path = self.base_folder / CFG_FILENAME
        self.google_client_secret_path = self.base_folder / CLIENT_SECRET_FILENAME

    def _ensure_base_folder(self):
        """Create base folder if not exists"""
        self.base_folder.mkdir(parents=True, exist_ok=True)

    def create_default(self):
        self._ensure_base_folder()
        with open(self.config_path, "w") as f:
            f.write(export_default_config())

    def read_config(self) -> Config:
        """ Read config file
            raise `CONFIG_NOT_EXISTS` if file not exists
        """
        if not self.config_path.exists():
            raise CONFIG_NOT_EXISTS(self.config_path)
        with open(self.config_path, "r") as f:
            cfg_toml = parse(f.read())
            return Config(**cfg_toml)

    def remove(self):
        if self.base_folder.exists:
            shutil.rmtree(self.base_folder)

    def get_base_path(self) -> Path:
        return self.base_folder



# custom_theme = Theme(
#     {"keyword": "bold magenta", "keyword2": "bold red", "safe": "green", "danger": "red"}
# )


def kw(text: str):
    return f"[keyword]{text}[/]"


def kw2(text: str):
    return f"[keyword2]{text}[/]"


def safe(text: str):
    return f"[safe]{text}[/]"


def danger(text: str):
    return f"[danger]{text}[/]"


# def get_console():
#     return Console(theme=custom_theme)
